# @(#)96        1.18  src/avs/fs/mmfs/rpm/SPECS/gpfs.gplbin.spec, mmfs, avs_rche, rches003a 11/13/09 13:30:04
#
# Licensed Materials - Property of IBM
#
# (C) COPYRIGHT International Business Machines Corp. 2005
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# Some helpful rpm packaging notes
#
# rpm does an upgrade/update in the following sequence
#
# 1) preinstall (%pre)
# 2) uninstall/install of files
# 3) postinstall (%post)
# 4) preuninstall (%preun)
# 5) postuninstall (%postun)
#
# The single argument $1 ($packageCnt) indicates the number of instances
# of the currently installed package.
#
# pre/post with $packageCnt==1 is a fresh install
# pre/post with $packageCnt==2 is an upgrade (over previous installed version)
# preun/postun with $packageCnt==0 is a full uninstall
# preun/postun with $packageCnt==1 is an upgrade uninstall (usually want to do nothing)
#

Summary: General Parallel File System
Name: gpfs.gplbin-%MODKERNEL
Version: 3.3.0
Release: 3
License: (C) COPYRIGHT International Business Machines Corp. 2009
Group: System Environment/Base
URL: http://www.ibm.com/servers/eserver/clusters/software/gpfs.html
Vendor: IBM Corp.
Packager: IBM Corp. <gpfs@us.ibm.com>

Autoreqprov : no
Requires: gpfs.base
%description
GPFS Compiled Open Source Modules

%install
  if [ x$RPM_BUILD_ROOT = x ]
  then
    echo "Build root not specified. Try --buildroot option on rpmbuild command"
    exit 1
  fi

  # Only perform file install when not in ODE environment
  if [ "$PROJECT_NAME" != "Linux" ]
  then
    # The following files and directories mimic our other non Linux
    # operating system install directories and modes.  Files that are
    # not shipped on Linux are commented out (not deleted) in an effort
    # to ease comparisons between different operating system install lists.
    #
    export GPFS_LPP=$RPM_BUILD_ROOT/usr/lpp/mmfs
    export GPFS_KO=$RPM_BUILD_ROOT/lib/modules/%MODKERNEL/extra
    export GPFS_BIN=$GPFS_LPP/bin
    export LPPBIN=$SHARKCLONEROOT/bin
  
    #############################################################
    # Following file list originated from AIX mmfs.base.rte.il  #
    #############################################################
    install -d -m 755 $GPFS_LPP
    install -d -m 755 $GPFS_BIN
    install -d -m 755 $GPFS_KO

    install -m 500    $LPPBIN/kdump-%MODKERNEL                     $GPFS_BIN/kdump-%MODKERNEL
    install -m 500    $LPPBIN/lxtrace-%MODKERNEL                   $GPFS_BIN/lxtrace-%MODKERNEL
    install -m 500    %GPLDIR/mmfs26.ko                  $GPFS_KO
    install -m 500    %GPLDIR/mmfslinux.ko               $GPFS_KO
    install -m 500    %GPLDIR/tracedev.ko                $GPFS_KO
  fi
  exit 0

%files

# NOTE: we specify file modes for ODE rpm packaging compatibility
#       because we skip the install phase in ODE (done w/ Makefiles).
# NOTE: all symbolic links and directories that are generated and not
#       in the shipping tree are handled in the post/postun sections.
# NOTE: Each file specifically has attributes assigned with %attr. This
#       has been chosen after prior maintainance with %defattr. It is
#       far simpler to maintain/compare multi-OS install lists in an
#       ordered fashion, which is not conducive to %defattr usage, other
#       than for directory creation and the config files section.

# Handle select directories first w/ default mode (755)
# Other directories are handled in post section to avoid
# having rpm try to automatically delete them.
%defattr(755,root,root)
%attr(500,root,root)  /usr/lpp/mmfs/bin/lxtrace-%MODKERNEL
%attr(500,root,root)  /usr/lpp/mmfs/bin/kdump-%MODKERNEL
%attr(500,root,root)  /lib/modules/%{MODKERNEL}/extra/mmfs26.ko
%attr(500,root,root)  /lib/modules/%{MODKERNEL}/extra/mmfslinux.ko
%attr(500,root,root)  /lib/modules/%{MODKERNEL}/extra/tracedev.ko
%defattr(-,root,root)

%pre
  #############################################################
  # Following .sh script is preinstallation script for gpfs   #
  #############################################################
  ##  Can not use rpm -qa | grep "mmfs.base.rte"  to check   ##
  ##  if mmfs is installed.                                  ##     
  #############################################################
  if test -n "$DEBUG" || test -n "$DEBUGpre"; then
    set -x
  fi
  packageCnt=$1

  # Handle upgrade case (if [ "$packageCnt" = 2 ])
  if test -x /usr/lpp/mmfs/bin/mmfsd && test -a /dev/null && test -z $GPFS_BG
  then
    # Unmount any existing GPFS file systems before installation
    /bin/umount -t gpfs > /dev/null 2>&1

    # Stop the MMFS before attempting installation
    if /bin/ps h -C mmfsd; then
      /usr/lpp/mmfs/bin/mmfsadm sd > /dev/null 2>&1
    fi

    # Remove the 'mmfs' kernel extension
    /usr/lpp/mmfs/bin/mmfsenv -u >/dev/null 2>&1
  fi

  exit 0

%post
  if test -n "$DEBUG" || test -n "$DEBUGpost"; then
    set -x
  fi
  packageCnt=$1

  # Most of these actions will be redundant for the
  # upgrade case (if [ "$packageCnt" = 2 ]), but this is the
  # easiest way to maintain the packaging when commands/links
  # are added and deleted in the future.

  # Make sure automatically generated path dirs
  # have correct permissions.  We must do this here
  # instead of %files to avoid having rpm spew out
  # annoying error messages.
  /bin/chmod 755 /usr/lpp
  /bin/chmod 755 /usr/lpp/mmfs
  /bin/chmod 755 /usr/lpp/mmfs/bin

  export GPFS_LPP=/usr/lpp/mmfs
  export GPFS_BIN=$GPFS_LPP/bin
  
  #do the depmod since we have already installed new KO's
  /sbin/depmod `uname -r`
  exit 0

%preun
  if test -n "$DEBUG" || test -n "$DEBUGpreun"; then
    set -x
  fi
  packageCnt=$1

  if [ "$packageCnt" = 0 ]; then
    # Uninstall

    # Skip cleanup if /dev/null does not exist or if GPFS_BG is set, since
    # that is a signal that the rpm is being installed in an alternate root
    # directory.
    if test -a /dev/null && test -z $GPFS_BG; then

      # Unmount any existing MMFS before installation
      /bin/umount -a -t gpfs >/dev/null 2>&1
      if [ $? -ne 0 ]
      then
        echo "Unable to unmount all GPFS file systems"
      fi

      # Stop the MMFS before if it is running
      if /bin/ps h -C mmfsd; then
        /usr/lpp/mmfs/bin/mmfsadm sd > /dev/null 2>&1
      fi

      # Remove the 'mmfs' kernel extension
      /usr/lpp/mmfs/bin/mmfsenv -u >/dev/null 2>&1
    fi
  fi 
  exit 0


%postun
  if test -n "$DEBUG" || test -n "$DEBUGpostun"; then
    set -x
  fi
  packageCnt=$1

  if [ "$packageCnt" = 0 ]; then
    # Uninstall
    export GPFS_LPP=/usr/lpp/mmfs
    export GPFS_BIN=$GPFS_LPP/bin

    /bin/rmdir  --ignore-fail-on-non-empty $GPFS_BIN

    # manually remove upper level dirs if empty
    # see corresponding note in %post for reasoning
    /bin/rmdir  --ignore-fail-on-non-empty /usr/lpp/mmfs
    /bin/rmdir  --ignore-fail-on-non-empty /usr/lpp
    #do the depmod since we have already installed new KO's
    /sbin/depmod `uname -r`

  else
    # Upgrade Uninstall

    # Any special upgrade uninstall actions go here
    :
  fi

  exit 0
