/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* @(#)41       1.108  src/avs/fs/mmfs/ts/kernext/gpl-linux/verdep.h, mmfs, avs_rche, rche0933a 7/29/09 19:26:19 */
#ifndef _h_verdep
#define _h_verdep

# include <linux/version.h>
# include <linux/buffer_head.h>
# include <linux/kdev_t.h>
# include <linux/namei.h>
# include <linux/mount.h>
# include <linux/page-flags.h>
# include <linux/security.h>
# include <linux/sched.h>
# include <linux/init.h>
# include <linux/fs.h>

# include <cxiTypes.h>
# include <cxiSystem.h>
# include <Logger-gpl.h>
# include <linux/moduleparam.h>

#if (LINUX_KERNEL_VERSION >= 2061200)
#include <linux/sort.h>
#endif
#if (defined(SUSE_LINUX) && (LINUX_KERNEL_VERSION == 2060507) && (!defined(I_WILL_FREE)))
#error "GPFS no longer supports SLES 9 distributions eariler than SP3."
#endif

/* from 2.6.22, thread_info is moved to stack in  task_struct */
#if (LINUX_KERNEL_VERSION >= 2062200)
#define TASK_THREAD_INFO(task) ((struct thread_info *)(task)->stack)
#else 
#define TASK_THREAD_INFO(task) (task->thread_info)
#endif 

/* Distro-specific defines */
# if (defined(SUSE_LINUX) && defined(LINUX_KERNEL_VERSION) && LINUX_KERNEL_VERSION >= 2062700)
#    define SUSE_SLES11  
# elif  (defined(SUSE_LINUX) && defined(SLE_VERSION_CODE) && SLE_VERSION_CODE >= 655872)
#    define SUSE_SLES10SP2
# endif
# if (defined(SUSE_LINUX) && defined(LINUX_KERNEL_VERSION) && LINUX_KERNEL_VERSION == 2060507)
#    define SUSE_SLES9
# endif
# if (defined(REDHAT_AS_LINUX) && defined(RHEL_RELEASE_CODE) && RHEL_RELEASE_CODE == 1282)
#    define REDHAT_RHEL52
#    define REDHAT_RHEL5
# endif
# if (defined(REDHAT_AS_LINUX) && defined(RHEL_RELEASE_CODE) && RHEL_RELEASE_CODE  == 1283)
#    define REDHAT_RHEL53
#    define REDHAT_RHEL5
# endif

/* Newer kernels fold readv and writev fops into aio_read and aio_write
   and use an iovec instead of buffer and length */
# if LINUX_KERNEL_VERSION >= 2061900
#   define KERNEL_HAS_NEW_AIO_FOPS
# endif

/* Newer kernels do not require work_struct data initialization */
# if LINUX_KERNEL_VERSION >= 2062000
# define GPFS_DECLARE_WORK(n, f) DECLARE_WORK(n, (work_func_t)f)
# else
# define GPFS_DECLARE_WORK(n, f) DECLARE_WORK(n, f, 0)
# endif

  /* Inode state indicators.
     SLES9 2.6.5-7.244+ kernels and SLES10 kernels require truncate_inode_pages
     to be called from the delete_inode sop and check for I_WILL_FREE flag, 
     RHEL4 kernels do not (thus far).  */
# if (defined(SUSE_LINUX) &&  LINUX_KERNEL_VERSION >= 2060507) || \
     (defined(REDHAT_AS_LINUX) && LINUX_KERNEL_VERSION > 2061800) || \
     (defined(KERNEL_ORG_LINUX) && LINUX_KERNEL_VERSION > 2061300)
     
#   define INODE_BEING_RELEASED (I_FREEING|I_CLEAR|I_WILL_FREE)
#   define INODE_INIT_INCOMPLETE (I_CLEAR|I_LOCK|I_NEW|I_WILL_FREE)
    /* required if file system defines delete_inode */
#   define MUST_TRUNCATE_INODE_PAGES 1
# else
#   define INODE_BEING_RELEASED (I_FREEING|I_CLEAR)
#   define INODE_INIT_INCOMPLETE (I_CLEAR|I_LOCK|I_NEW)
#   define MUST_TRUNCATE_INODE_PAGES 0
# endif

/* Sort routine not provided in all kernels */
# if defined(REDHAT_LINUX) || (LINUX_KERNEL_VERSION >= 2061200)
#   include <linux/sort.h>
#   define SORT(a,b,s,fn)   sort(a,b,s,fn,NULL)
# elif defined(SUSE_SLES9)
#   define SORT(a,b,s,fn)   qsort(a,b,s,fn)
# else
#   define SORT(a,b,s,fn)   do {} while (0)
/*  #error "No sort routine defined for this kernel" */
# endif


  /* Support for cluster-aware NFS locks in present in kernels > 2.6.22
     and backported to SLES10 SP2 and RHEL > 5.2 */
# if defined(SUSE_SLES10SP2) || defined(REDHAT_RHEL5) || \
     (LINUX_KERNEL_VERSION >= 2062200)
#   define NFS_CLUSTER_LOCKS
# endif

  /* Support for cluster-aware leases */
# if defined(SUSE_SLES10SP2) || defined(REDHAT_RHEL53) || \
     (LINUX_KERNEL_VERSION >= 2062300)
#   define CLUSTER_LEASES
# endif

  /* Call to issue POSIX setlease */
# ifdef CLUSTER_LEASES
#   ifdef REDHAT_RHEL53
#     define POSIX_SETLEASE(fP, arg, flPP)  __setlease(fP, arg, flPP)
#   else
#     define POSIX_SETLEASE(fP, arg, flPP)  generic_setlease(fP, arg, flPP)
#   endif
# else
#     define POSIX_SETLEASE(fP, arg, flPP)  0
# endif

  /* Uncache a directory entry.  Must hold dcache lock. */
# define DENTRY_DROP(DP)  \
  {  \
    spin_lock(&(DP)->d_lock); \
    __d_drop(DP); \
    spin_unlock(&(DP)->d_lock); \
  }

  /* Uncache a directory entry.  Must hold *NO* locks. */
  /* __d_drop checks state of DCACHE_UNHASHED prior to drop. */
# define DENTRY_D_DROP(DP)  \
  {  \
    spin_lock(&dcache_lock); \
    spin_lock(&(DP)->d_lock); \
    __d_drop(DP); \
    spin_unlock(&(DP)->d_lock); \
    spin_unlock(&dcache_lock); \
  }


#if LINUX_KERNEL_VERSION >= 2061300 || \
   (LINUX_KERNEL_VERSION == 2060507 && defined(SUSE_LINUX))
#define MY_INIT_LIST_HEAD(SINODES) INIT_LIST_HEAD(SINODES)
#else
#define MY_INIT_LIST_HEAD(SINODES) 
#endif

#if (LINUX_KERNEL_VERSION >= 2062400)
# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
    (SBP)->s_flags |= MS_ACTIVE;                  \
    MY_INIT_LIST_HEAD(&((SBP)->s_inodes));        \
  }

# define UNSET_SUPER_BLOCK(SBP)                   \
  {						  \
  }
#else
# define SET_SUPER_BLOCK(SBP,SOPS)                \
  {                                               \
    memset((SBP), 0, sizeof(struct super_block)); \
    (SBP)->s_op = SOPS;                           \
    (SBP)->s_flags |= MS_ACTIVE;                  \
    security_sb_alloc(SBP);			  \
    MY_INIT_LIST_HEAD(&((SBP)->s_inodes));        \
  }

# define UNSET_SUPER_BLOCK(SBP)              \
  {						  \
    security_sb_free(SBP);			  \
  }
#endif

#if (LINUX_KERNEL_VERSION >= 2062500)
#define sb_entry(list)        list_entry((list), struct super_block, s_list)
#define TASK_PWD(tsP)         ((tsP)->fs ? &(tsP)->fs->pwd : NULL)
#define TASK_PWDMNT(tsP)      ((tsP)->fs ? (tsP)->fs->pwd.mnt : NULL)
#define TASK_ROOTMNT(tsP)     ((tsP)->fs ? (tsP)->fs->root.mnt : NULL)
#define NDP_TO_IP(ndP)        (ndP)->path.dentry->d_inode;
#define ND_PATH(nd)           (nd).path
#define EXP_DENTRY(expP)      (expP)->ex_path.dentry
#define PATH_PUT(ndP)         path_put(&(ndP)->path)
#else
#define TASK_PWD(tsP)         ((tsP)->fs ? (tsP)->fs->pwd : NULL)
#define TASK_PWDMNT(tsP)      ((tsP)->fs ? (tsP)->fs->pwdmnt : NULL)
#define TASK_ROOTMNT(tsP)     ((tsP)->fs ? (tsP)->fs->rootmnt : NULL)
#define NDP_TO_IP(ndP)        (ndP)->dentry->d_inode
#define ND_PATH(nd)           (nd)
#define EXP_DENTRY(expP)      (expP)->ex_dentry
#define PATH_PUT(ndP)         path_release(ndP)
#endif

#if (LINUX_KERNEL_VERSION >= 2062400)
# define PROCESS_GROUP(T)            task_pgrp_nr(T)
#else
# define PROCESS_GROUP(T)            process_group(T)
#endif


#if defined(NFS_CLUSTER_LOCKS) && !defined(REDHAT_RHEL5)
# define POSIX_LOCK_FILE(FLP1,FLP2)  posix_lock_file((FLP1),(FLP2), NULL)
# define posix_lock_file_conf        posix_lock_file
#else
# define POSIX_LOCK_FILE(FLP1,FLP2)  posix_lock_file((FLP1),(FLP2))
#endif

# define CXITIME_TO_INODETIME(CXITIME, ITIME)   \
  {                                             \
    (ITIME).tv_sec = (CXITIME).tv_sec;          \
    (ITIME).tv_nsec = (CXITIME).tv_nsec;        \
  }
# define CXITIME_FROM_INODETIME(CXITIME, ITIME) \
  {                                             \
    (CXITIME).tv_sec = (ITIME).tv_sec;          \
    (CXITIME).tv_nsec = (ITIME).tv_nsec;        \
  }

#if (LINUX_KERNEL_VERSION < 2061700)
#define NFS_FATTR_FSID(fattrP)  ((fattrP)->fsid_u.nfs4)
#else
#define HAS_NFS_SUBMOUNTS
#define NFS_FATTR_FSID(fattrP)  ((fattrP)->fsid)
#endif

#if (LINUX_KERNEL_VERSION >= 2061400)
/* nd->intent.open */
#define LOOKUP_INTENT
#endif

#ifdef SUSE_SLES9
#define INTENT_OPEN_FLAGS(intent)       (intent).it_flags
#define INTENT_OPEN_CREATE_MODE(intent) (intent).it_create_mode
#else
#define INTENT_OPEN_FLAGS(intent)       (intent).open.flags
#define INTENT_OPEN_CREATE_MODE(intent) (intent).open.create_mode
#endif

#if (LINUX_KERNEL_VERSION >= 2061300)
#define NFS_CACHE_VALIDITY_FLAGS(iP)  NFS_I(iP)->cache_validity
#else
#define NFS_CACHE_VALIDITY_FLAGS(iP)  NFS_I(iP)->flags
#endif

/* module_param_string argument setting the permissions on
   prog_path and mmfsd_path, in /sys/module/mmfslinux/parameters */
#define PERM 0664

#if (LINUX_KERNEL_VERSION >= 2060700)
#define PUTNAME __putname
#else
#define PUTNAME putname
#endif

#ifdef GPFS_ARCH_IA64
   /* user addresses are below region 5 (while PAGE_OFFSET is at region 7) */
#define GPFS_KERNEL_OFFSET 0xa000000000000000UL
#elif defined( GPFS_ARCH_I386)
#define GPFS_KERNEL_OFFSET (UIntPtr) PAGE_OFFSET 
#else
#define GPFS_KERNEL_OFFSET (UIntPtr) PAGE_OFFSET
#endif /* GPFS_ARCH_IA64 */

#if LINUX_KERNEL_VERSION >= 2060600 || \
   (LINUX_KERNEL_VERSION == 2060507 && defined(SUSE_LINUX))
#define UNMAP_MAPPING_RANGE(MAPPING, START, STOP) unmap_shared_mapping_range(MAPPING, START, STOP);
#define MMAP_WRITE_ACCESS(WRITEACCESS) WRITEACCESS = true;
#elif defined(MMAP_LINUX_PATCH)
#define UNMAP_MAPPING_RANGE(MAPPING, START, STOP) invalidate_mappings(MAPPING, START);
#define MMAP_WRITE_ACCESS(WRITEACCESS) WRITEACCESS = true;
#else
#define UNMAP_MAPPING_RANGE(MAPPING, START, STOP) 
#define MMAP_WRITE_ACCESS(WRITEACCESS) WRITEACCESS = false;
#endif

#if LINUX_KERNEL_VERSION >= 2060900
#define VFS_FOLLOW_LINK(RC, ND, BUF) nd_set_link(ND, BUF); BUF = NULL
#else
#define VFS_FOLLOW_LINK(RC, ND, BUF) RC = vfs_follow_link(ND, BUF)
#endif

#ifdef SUSE_SLES10SP2
#define VFS_MKDIR(vfsP, diP, dP, mode)        vfs_mkdir(diP, dP, vfsP, mode)
#define VFS_MKNOD(vfsP, diP, dP, mode, dev)   vfs_mknod(diP, dP, vfsP, mode, dev)
#define VFS_UNLINK(vfsP, diP, dP)             vfs_unlink(diP, dP, vfsP)
#define VFS_RMDIR(vfsP, diP, dP)              vfs_rmdir(diP, dP, vfsP)
#define VFS_LINK(vfsP, sdP, diP, tdP)         vfs_link(sdP, vfsP, diP, tdP, vfsP)
#define VFS_RENAME(vfsP, sdiP, sdP, tdiP, tdP) vfs_rename(sdiP, sdP, vfsP, tdiP, tdP, vfsP);
#elif defined SUSE_SLES11
#define VFS_MKDIR(vfsP, diP, dP, mode)        vfs_mkdir(diP, dP, vfsP, mode)
#define VFS_MKNOD(vfsP, diP, dP, mode, dev)   vfs_mknod(diP, dP, vfsP, mode, dev)
#define VFS_UNLINK(vfsP, diP, dP)             vfs_unlink(diP, dP, vfsP)
#define VFS_RMDIR(vfsP, diP, dP)              vfs_rmdir(diP, dP, vfsP)
#define VFS_LINK(vfsP, sdP, diP, tdP)         vfs_link(sdP, vfsP, diP, tdP, vfsP)
#define VFS_RENAME(vfsP, sdiP, sdP, tdiP, tdP) vfs_rename(sdiP, sdP, vfsP, tdiP, tdP, vfsP);
#else
#define VFS_MKDIR(vfsP, diP, dP, mode)        vfs_mkdir(diP, dP, mode)
#define VFS_MKNOD(vfsP, diP, dP, mode, dev)   vfs_mknod(diP, dP, mode, dev)
#define VFS_UNLINK(vfsP, diP, dP)             vfs_unlink(diP, dP)
#define VFS_RMDIR(vfsP, diP, dP)              vfs_rmdir(diP, dP)
#define VFS_LINK(vfsP, sdP, diP, tdP)         vfs_link(sdP, diP, tdP)
#define VFS_RENAME(vfsP, sdiP, sdP, tdiP, tdP) vfs_rename(sdiP, sdP, tdiP, tdP);
#endif

#if (!defined(SUSE_SLES11) && LINUX_KERNEL_VERSION >= 2062700) || defined(SUSE_SLES9)
#define VFS_SYMLINK(vfsP, diP, dP, nameP)     vfs_symlink(diP, dP, nameP);
#elif defined(SUSE_SLES10SP2) 
#define VFS_SYMLINK(vfsP, diP, dP, nameP)     vfs_symlink(diP, dP, vfsP, nameP, 0);
#elif defined(SUSE_SLES11)
#define VFS_SYMLINK(vfsP, diP, dP, nameP)     vfs_symlink(diP, dP, vfsP, nameP);
#else
#define VFS_SYMLINK(vfsP, diP, dP, nameP)     vfs_symlink(diP, dP, nameP, 0);
#endif

#if LINUX_KERNEL_VERSION >= 2061000
#define MY_RLIM_CUR(WHAT) current->signal->rlim[WHAT].rlim_cur
#else
#define MY_RLIM_CUR(WHAT) current->rlim[WHAT].rlim_cur
#endif

#if LINUX_KERNEL_VERSION >= 2061000
#define GENERIC_PERMISSION gpfs_i_permission_noacl
#else
#define GENERIC_PERMISSION vfs_permission
#endif

#if LINUX_KERNEL_VERSION >= 2061600
#define d_child d_u.d_child
#endif

#if LINUX_KERNEL_VERSION >= 2060900
#define HAS_IOP_PUT_LINK
#endif

#if (LINUX_KERNEL_VERSION > 2061800) || defined(REDHAT_RHEL5)
#define PRVINODE i_private
#else
#define PRVINODE u.generic_ip
#endif

#define cxiDevMajor(D)       ((UInt32)((D) >> 20))
#define cxiDevMinor(D)       ((UInt32)((D) & 0xfffff))
#define cxiDevToKernelDev(D) new_decode_dev(D)
#define cxiDev32ToDev(D) \
  (cxiDev_t)((cxiDev32Major(D) << 8) | \
             (cxiDev32Minor(D) & 0xFF) | \
             ((cxiDev32Minor(D) & ~0xFF) << 12))

#if LINUX_KERNEL_VERSION >= 2062000
#define SLAB_KERNEL GFP_KERNEL
#endif

#if LINUX_KERNEL_VERSION < 2061600
#define INODE_LOCK(iP)    down(&(iP)->i_sem)
#define INODE_UNLOCK(iP)  up(&(iP)->i_sem)
#else
#define INODE_LOCK(iP)    mutex_lock(&(iP)->i_mutex)
#define INODE_UNLOCK(iP)  mutex_unlock(&(iP)->i_mutex)
#endif

#if LINUX_KERNEL_VERSION < 2061800
#define HAS_TASKLIST_LOCK
#define TASK_READ_LOCK read_lock(&tasklist_lock);
#define TASK_READ_UNLOCK read_unlock(&tasklist_lock);
#else
#include <linux/rcupdate.h>
#endif


#if LINUX_KERNEL_VERSION >= 2061900 || \
    (defined(REDHAT_AS_LINUX) && LINUX_KERNEL_VERSION >= 2061800)
#define INODE_USES_BLKBITS
#endif 

#if defined(INODE_USES_BLKBITS)
#define SET_INODE_BLOCKSIZE(inodeP, blocksize) \
  (inodeP)->i_blkbits = myblksize_bits((unsigned int)(blocksize));
#define GET_INODE_BLOCKSIZE(inodeP) (1 << (inodeP)->i_blkbits)
#else
#define SET_INODE_BLOCKSIZE(inodeP, blocksize) \
  (inodeP)->i_blksize = (blocksize);
#define GET_INODE_BLOCKSIZE(inodeP) ((inodeP)->i_blksize)
#endif

#if LINUX_KERNEL_VERSION < 2061700
#define HAS_PIDTYPE_TGID
#endif

#if LINUX_KERNEL_VERSION >= 2062000
#define KMEM_CACHE_T struct kmem_cache
#else
#define KMEM_CACHE_T kmem_cache_t
#endif

#if LINUX_KERNEL_VERSION >= 2062600
#define SEM_COUNT(semP)         (semP)->count
#define SEM_WAIT_TASKLIST(semP) ((semP)->wait_list)
#else
#define SEM_COUNT(semP)         atomic_read(&(semP)->count)
#define SEM_WAIT_TASKLIST(semP) ((semP)->wait.task_list)
#endif

#if LINUX_KERNEL_VERSION >= 2062700
#define FIND_TASK_BY_PID find_task_by_vpid
typedef struct request_queue request_queue_t;
#define USER_PATH(pathname, p) user_path(pathname, &(p)->path)
#define USER_LPATH(pathname, p) user_lpath(pathname, &(p)->path)
#define LONG_FILECOUNT
#else
#define FIND_TASK_BY_PID find_task_by_pid
#define USER_PATH(pathname, p) user_path_walk(pathname, p)
#define USER_LPATH(pathname, p) user_path_walk_link(pathname, p)
#endif
#if LINUX_KERNEL_VERSION >= 2062800
#define d_alloc_anon d_obtain_alias
#endif
#if LINUX_KERNEL_VERSION >= 2062900
#define CRED(ctP, fl)         (ctP)->cred->fl
#define CRED_GRP(ctP, fl)     (ctP)->cred->group_info->fl
#define DENTRY_OPEN(dP, mntP, flags) dentry_open(dP, mntP, flags, current->cred)
#else
#if LINUX_KERNEL_VERSION > 2060300
#define CRED_GRP(ctP, fl)     (ctP)->group_info->fl
#else
#define CRED_GRP(ctP, fl)     (ctP)->fl
#endif
#define CRED(ctP, fl)         (ctP)->fl
#define DENTRY_OPEN(dP, mntP, flags) dentry_open(dP, mntP, flags)
#endif

/* Translate to GPFS native disk format which used high and low 16 bits
 * for major/minor
 */
#define cxiDevToDev32(D) \
  (cxiDev32_t)((cxiDevMajor(D) << 16) | (cxiDevMinor(D) & 0xFFFF))

#define MAX_GPFS_KMALLOC (131072)

#if LINUX_KERNEL_VERSION < 2062100
#define RELAY_OPEN(cpu, dir, size, num, cb) relay_open(cpu, dir, size, num, cb)
#else
#define RELAY_OPEN(cpu, dir, size, num, cb) relay_open(cpu, dir, size, num, cb, NULL)
#endif

#endif /* _h_verdep */
